@@ -13,8 +13,11 @@
 import logging
 import sys
 import textwrap
+import warnings
+import yaml
 
 from oslo_config import cfg
+from oslo_serialization import jsonutils
 import stevedore
 
 from oslo_policy import policy
@@ -44,6 +47,12 @@ ENFORCER_OPTS = [
                     'which to look for a policy.Enforcer.'),
 ]
 
+UPGRADE_OPTS = [
+    cfg.StrOpt('policy',
+               required=True,
+               help='Path to the policy file which need to be updated.')
+]
+
 
 def get_policies_dict(namespaces):
     """Find the options available via the given namespaces.
@@ -89,19 +98,45 @@ def _format_help_text(description):
     if not description:
         return '#'
 
-    lines = description.splitlines()
     formatted_lines = []
-    # wrap each line to support multi line descriptions
-    for line in lines:
-        if not line:
+    paragraph = []
+
+    def _wrap_paragraph(lines):
+        return textwrap.wrap(' '.join(lines), 70, initial_indent='# ',
+                             subsequent_indent='# ')
+
+    for line in description.strip().splitlines():
+        if not line.strip():
+            # empty line -> line break, so dump anything we have
+            formatted_lines.extend(_wrap_paragraph(paragraph))
             formatted_lines.append('#')
+            paragraph = []
+        elif len(line) == len(line.lstrip()):
+            # no leading whitespace = paragraph, which should be wrapped
+            paragraph.append(line.rstrip())
         else:
-            formatted_lines.append(textwrap.fill(line, 70,
-                                   initial_indent='# ',
-                                   subsequent_indent='# ',
-                                   break_long_words=False,
-                                   replace_whitespace=False))
-    return "\n".join(formatted_lines)
+            # leading whitespace - literal block, which should not be wrapping
+            if paragraph:
+                # ...however, literal blocks need a new line before them to
+                # delineate things
+                # TODO(stephenfin): Raise an exception here and stop doing
+                # anything else in oslo.policy 2.0
+                warnings.warn(
+                    'Invalid policy description: literal blocks must be '
+                    'preceded by a new line. This will raise an exception in '
+                    'a future version of oslo.policy:\n%s' % description,
+                    FutureWarning)
+                formatted_lines.extend(_wrap_paragraph(paragraph))
+                formatted_lines.append('#')
+                paragraph = []
+
+            formatted_lines.append('# %s' % line.rstrip())
+
+    if paragraph:
+        # dump anything we might still have in the buffer
+        formatted_lines.extend(_wrap_paragraph(paragraph))
+
+    return '\n'.join(formatted_lines)
 
 
 def _format_rule_default_yaml(default, include_help=True):
@@ -303,6 +338,48 @@ def generate_policy(args=None):
     _generate_policy(conf.namespace, conf.output_file)
 
 
+def _upgrade_policies(policies, default_policies):
+    old_policies_keys = list(policies.keys())
+    for section in sorted(default_policies.keys()):
+        rule_defaults = default_policies[section]
+        for rule_default in rule_defaults:
+            if (rule_default.deprecated_rule and
+                    rule_default.deprecated_rule.name in old_policies_keys):
+                policies[rule_default.name] = policies.pop(
+                    rule_default.deprecated_rule.name)
+                LOG.info('The name of policy %(old_name)s has been upgraded to'
+                         '%(new_name)',
+                         {'old_name': rule_default.deprecated_rule.name,
+                          'new_name': rule_default.name})
+
+
+def upgrade_policy(args=None):
+    logging.basicConfig(level=logging.WARN)
+    conf = cfg.ConfigOpts()
+    conf.register_cli_opts(GENERATOR_OPTS + RULE_OPTS + UPGRADE_OPTS)
+    conf.register_opts(GENERATOR_OPTS + RULE_OPTS + UPGRADE_OPTS)
+    conf(args)
+    with open(conf.policy, 'r') as input_data:
+        policies = policy.parse_file_contents(input_data.read())
+    default_policies = get_policies_dict(conf.namespace)
+
+    _upgrade_policies(policies, default_policies)
+
+    if conf.output_file:
+        if conf.format == 'yaml':
+            yaml.safe_dump(policies, open(conf.output_file, 'w'),
+                           default_flow_style=False)
+        elif conf.format == 'json':
+            jsonutils.dump(policies, open(conf.output_file, 'w'),
+                           indent=4)
+    else:
+        if conf.format == 'yaml':
+            sys.stdout.write(yaml.safe_dump(policies,
+                                            default_flow_style=False))
+        elif conf.format == 'json':
+            sys.stdout.write(jsonutils.dumps(policies, indent=4))
+
+
 def list_redundant(args=None):
     logging.basicConfig(level=logging.WARN)
     conf = cfg.ConfigOpts()
